/* ---------------------------------------------------------------------------- */
/* AbstractSurface                                                              */
/* ---------------------------------------------------------------------------- */

/* Fonctionnalités et code communs aux Surfaces.                                */
/* Pour éviter quelques duplications de code, il est possible de faire hériter  */
/* une nouvelle surface de la classe AbstractSurface.                           */
/* Les fonctionnalités des surfaces codées ici sont principalement celles qui   */
/* touchent à la gestion de données de la surface :                             */
/* * manipulation des positions                                                 */
/* * affichage des faces                                                        */
/* * déformation de la surface par un flot (ArrayList<PVector>)                 */
/* Pour la gestion de l'adjacence des positions sur la surface il est conseillé */
/* d'utiliser une structure de données plus évoluée. (liste ou matrice          */
/* d'adjacence par exemple)                                                     */
/* À noter que les données sont partagées car non dupliquées : si une structure */
/* de données peut être plus efficace pour un calcul il est possible de         */
/* l'utiliser pour ce calcul et impacter l'autre surface. (possible mais astuce */
/* non utilisée dans ce projet)                                                 */

abstract class AbstractSurface implements Surface {
  SurfaceData data;
  private PVectorList flat_vertex_normals = null;
  private PVector barycenter = null;
  
  AbstractSurface(SurfaceData data) {
    this.data = data;
    this.flat_vertex_normals = new PVectorList();
    for(int i = 0; i < data.faces.size(); ++i)
      this.flat_vertex_normals.add(new PVector(0, 0, 0));
  }
  
  @Override
  SurfaceData get_data() {
    return data;
  }
  
  @Override
  PVector get(int vertex_id) {
    return data.vertices.get(vertex_id);
  }
  
  @Override
  PVector get(int first, int second) {
    return PVector.sub(get(second), get(first));
  }
  
  @Override
  IntList get_face(int f_id) {
    return data.faces.get(f_id);
  }
  
  @Override
  Iterable<PVector> vertices() {
    return data.vertices;
  }
  
  @Override
  Iterable<IntList> faces() {
    return data.faces;
  }
  
  @Override
  Iterable<Integer> vertices_ids() {
    return new IntRange(nV());
  }
  
  @Override
  Iterable<Integer> faces_ids() {
    return new IntRange(nF());
  }
  
  @Override
  int nV() {
    return data.vertices.size();
  }
  
  int nF() {
    return data.faces.size();
  }
  
  @Override
  ArrayList<PVector> vectors_from_ids(IntList ids) {
    ArrayList<PVector> res = new ArrayList<PVector>();
    for(int id : ids)
      res.add(get(id));
    return res;
  }
  
  private float face_area(IntList face) {
    float area = 0;
    for(int i = 2; i < face.size(); ++i) {
      area += get(face.get(0), face.get(i - 1)).cross(get(face.get(0), face.get(i))).mag() / 2;
    }
    return area;
  }
  
  private boolean face_is_drawable(IntList face) {
    if(abs(face_area(face)) < 1e-6)
      return false;
    return true;
  }
  
  @Override
  void draw_from_ground_truth_face_error(SurfaceData ground_truth) {
    if(ground_truth.vertices.size() != get_data().vertices.size())
      throw new IllegalArgumentException("We must have a correspondance between the surface and its ground truth");
    float max_error = Float.MIN_VALUE;
    float min_error = Float.MAX_VALUE;
    float error;
    
    ArrayList<FloatList> face_vertices_errors = new ArrayList<FloatList>();
    for(IntList face : faces()) { /* Ici on regarde la déformation de chaque face par rapport à une face de référence */
      PVectorList face_this = new PVectorList(vectors_from_ids(face));
      PVectorList face_other = ground_truth.vertices.vectors_from_ids(face);
      PVector center_this = face_this.barycenter();
      PVector center_other = face_other.barycenter();
      FloatList res = new FloatList();
      for(int i = 0; i < face.size(); ++i) { /* On compare les distances entre les vecteur du barycentre de la face au sommet de celle-ci */
        res.append(error = PVector.sub(PVector.sub(face_this.get(i), center_this), PVector.sub(face_other.get(i), center_other)).mag());
        max_error = max(max_error, error);
        min_error = min(min_error, error);
      }
      face_vertices_errors.add(res);
    }
    
    PVector p;
    
    for(int f_id : faces_ids()) { /* On dessine chaque face et une couleur représentative de la déformation des sommêts sur cette face */
      IntList f = get_face(f_id);
      if(! face_is_drawable(f))
        continue;
      beginShape();
      for(int i = 0; i < f.size(); ++i) {
        int v_id = f.get(i);
        int vcolor;
        if(abs(max_error - min_error) < 1e-9)
          vcolor = 0;
        else
          vcolor = ((int)(255 * pow((max_error - face_vertices_errors.get(f_id).get(i)) / (max_error - min_error), 2)));
        fill(255 - vcolor, vcolor, 0);
        stroke((255 - vcolor) * 0.7, vcolor * 0.7, 0);
        p = get_data().vertices.get(v_id);
        vertex(p.x, p.y, p.z);
      }
      endShape();
    }
    stroke(250);
    fill(127);
  }
  
  @Override
  void draw_from_ground_truth(SurfaceData ground_truth) {
    if(ground_truth.vertices.size() != get_data().vertices.size()) {
      /*throw new IllegalArgumentException("We must have a correspondance between the surface and its ground truth");*/
      draw();
      return;
    }
    float max_error = Float.MIN_VALUE;
    float min_error = Float.MAX_VALUE;
    float error;
    
    FloatList vertices_errors = new FloatList();
    for(int v_id : vertices_ids()) { /* On compare la distance de chaque sommêt au sommêt dans la surface de référence */
      vertices_errors.append(error = PVector.sub(get(v_id), ground_truth.vertices.get(v_id)).mag());
      max_error = max(max_error, error);
      min_error = min(min_error, error);
    }
    
    PVector p;
    
    for(int f_id : faces_ids()) { /* On dessine chaque face et chaque sommêt avec la couleur qui correspond à son erreur au sommêt du ground truth */
      IntList f = get_face(f_id);
      if(! face_is_drawable(f))
        continue;
      beginShape();
      for(int i = 0; i < f.size(); ++i) {
        int v_id = f.get(i);
        int vcolor;
        if(abs(max_error - min_error) < 1e-9)
          vcolor = 0;
        else
          vcolor = ((int)(255 * ((max_error - vertices_errors.get(v_id)) / (max_error - min_error)))); /* renormalisation qui envoie une erreur de [min, max] sur [[0, 255]] */
        fill(255 - vcolor, vcolor, 0);
        stroke((255 - vcolor) * 0.7, vcolor * 0.7, 0);
        p = get(v_id);
        vertex(p.x, p.y, p.z);
      }
      endShape();
    }
    stroke(250);
    fill(127);
  }
  
  @Override
  void draw() {
    PVector p;
    if(lines_on_surface)
      stroke(250);
    else
      noStroke();
    if(grid) {
      stroke(255, 216, 0);
      noFill();
      for(IntList f : faces()) {
        if(! face_is_drawable(f))
          continue;
        for(int i = 0; i < f.size(); ++i) {
          p = get(f.get(i));
          PVector q = get(f.get((i + 1) % f.size()));
          line(p.x, p.y, p.z, q.x, q.y, q.z);
        }
      }
    }
    else {
      fill(0x5F, 0x6B, 0x81);
      for(IntList f : faces()) {
        if(! face_is_drawable(f))
          continue;
        beginShape();
        for(int v_id : f) {
          p = get(v_id);
          vertex(p.x, p.y, p.z);
        }
        endShape();
      }
    }
  }
  
  @Override
  void draw_flow(ArrayList<PVector> vector_field, float tau) {
    if(other_style)
      draw_flow(vector_field, tau, 0, 0, 0);
    else
      draw_flow(vector_field, tau, 255, 255, 255);
  }
  
  void draw_flow(ArrayList<PVector> vector_field, float tau, int r, int g, int b)  {
    stroke(r, g, b);
    PVector p;
    PVector q;
    for(int i = 0; i < vector_field.size(); ++i) {
      p = get(i);
      q = PVector.add(p, PVector.mult(vector_field.get(i), tau));
      line(p.x, p.y, p.z, q.x, q.y, q.z);
    }
  }
  
  private PVector face_barycenter(IntList face) {
    PVector p = new PVector(0, 0, 0);
    for(int v_id : face)
      p.add(get(v_id));
    p.div(face.size());
    return p;
  }
  
  @Override
  void draw_faces_normals(float size) {
    stroke(0, 216, 255);
    ArrayList<PVector> face_normals = new ArrayList<PVector>();
    for(int f_id : faces_ids()) {
      face_normals.add(face_normal(f_id));
    }
    for(int f_id : faces_ids()) {
      IntList face = get_face(f_id);
      PVector p = face_barycenter(face);
      PVector q = PVector.add(p, PVector.mult(face_normals.get(f_id), size));
      line(p.x, p.y, p.z, q.x, q.y, q.z);
    }
  }
  
  @Override
  void scale(float scalar) {
    for(PVector v : vertices())
      v.mult(scalar);
  }
  
  @Override
  void translate(PVector direction) {
    for(PVector v : vertices())
      v.add(direction);
  }
  
  @Override
  PVector barycenter() {
    PVector p = new PVector(0, 0, 0);
    for(PVector v : vertices())
      p.add(v);
    p.div(nV());
    return p;
  }
  
  float star_area(int p) {
    float area = 0;
    IntList star = orientedNeighbors(p);
    for(int i = 0; i < star.size(); ++i)
      area += get(p, star.get(i)).cross(get(p, star.get((i + 1) % star.size()))).mag() / 2;
    return area;
  }
  
  float area() {
    float area = 0;
    for(IntList face : faces()) {
      for(int i = 2; i < face.size(); ++i) {
        area += get(face.get(0), face.get(i - 1)).cross(get(face.get(0), face.get(i))).mag() / 2;
      }
    }
    return area;
  }
  
  float volume() {
    float volume = 0;
    PVector centroid = barycenter();
    for(IntList face : faces()) {
      PVector v1 = PVector.sub(get(face.get(0)), centroid);
      for(int i = 2; i < face.size(); ++i) {
        PVector v2 = PVector.sub(get(face.get(i - 1)), centroid);
        PVector v3 = PVector.sub(get(face.get(i)), centroid);
        volume += v1.dot(v2.cross(v3)) / 6;
      }
    }
    return abs(volume);
  }
  
  @Override
  void auto_scale(float max_size) {
    PVector center = barycenter();
    float max_length = 0;
    for(PVector v : vertices())
      max_length = max(PVector.sub(center, v).mag(), max_length);
    scale(max_size / max_length);
  }
  
  @Override
  void flow(ArrayList<PVector> vector_field, float tau) {
    for(int i : vertices_ids()) {
      get(i).add(PVector.mult(vector_field.get(i), tau));
    }
  }
  
  @Override
  abstract IntList orientedNeighbors(int p); 
  
  @Override
  float defaut_angulaire(int p) {
    IntList voisins = orientedNeighbors(p);
    float theta = 0;
    for(int i = 0; i < voisins.size(); ++i)
      theta += PVector.angleBetween(get(p, voisins.get(i)), get(p, voisins.get((i + 1) % voisins.size())));
    return TWO_PI - theta;
  }
  
  void precompute_barycenter_to_ajust_normals() {
    barycenter = barycenter();
  }
  
  void reset_precomputed_barycenter() {
    barycenter = null;
  }
  
  @Override
  PVector face_normal(int f) {
    PVector res;
    IntList face = get_face(f);
    int i1 = 1;
    int i2 = 2;
    for(; i1 < face.size(); ++i1)
      if(norme_carre(get(face.get(i1), face.get(0))) > 1e-6)
        break;
    for(i2 = i1 + 1; i2 < face.size(); ++i2)
      if(norme_carre(get(face.get(i2), face.get(0))) > 1e-6 && norme_carre(PVector.sub(get(face.get(i1), face.get(0)), get(face.get(i2), face.get(0)))) > 1e-6)
        break;
    if(i2 < face.size()) {
      res = get(face.get(i1), face.get(0)).cross(get(face.get(i2), face.get(0)));
    } else
      res = flat_vertex_normals.get(f);
    res.normalize();
    if(barycenter != null) {
      if(PVector.sub(face_barycenter(face), barycenter).dot(res) < 0)
        res.mult(-1);
    }
    return res;
  }
  
  /* vérifie si la surface est orientée de sorte que la normale des faces soit orientée vers l'extérieur et change l'orientation si nécessaire */
  void check_and_change_orientation() {
    int ok = 0;
    int bad = 0;
    PVector centroid = barycenter();
    for(int f_id : faces_ids()) {
      IntList face = get_face(f_id);
      PVector tmp = new PVector(0, 0, 0);
      for(int v : face)
        tmp.add(get(v));
      tmp.div(face.size());
      if(PVector.dot(face_normal(f_id), PVector.sub(centroid, tmp)) > 0)
        ++bad;
      else
        ++ok;
    }
    if(bad > ok)
      for(IntList face : faces())
        face.reverse();
  }
  
  PVectorList faces_normal() {
    PVectorList res = new PVectorList();
    for(int f_id : faces_ids())
      res.add(face_normal(f_id));
    return res;
  }
  
  private IntList rebuild_face(int v_id, IntList face, IntList neighbors, IntList new_face) {
    IntList res = new IntList();
    for(int i = 0; i < face.size(); ++i) {
      if(face.get(i) == v_id) { /* on remplace p par une arête sur la face obtenue par son écrasement */
        for(int j = 0; j < neighbors.size(); ++j) {
          if(neighbors.get(j) != face.get((i + 1) % face.size()))
            continue;
          if(neighbors.get((j + 1) % neighbors.size()) == face.get((i - 1 + face.size()) % face.size())) { /* on refabrique la face telle qu'on garde l'orientation de la surface */
            res.append(new_face.get((j + 1) % new_face.size()));
            res.append(new_face.get(j));
          } else {
            res.append(new_face.get(j));
            res.append(new_face.get((j + 1) % new_face.size()));
          }
          break;
        }
      } else
        res.append(face.get(i));
    }
    return res;
  }
  
  @Override
  int flat(int v_id, PVector vertex_normal) {
    int degree = vertex_degree(v_id);
    IntList neighbors = orientedNeighbors(v_id);
    IntList faces = faces_in(v_id);
    PVector p = get(v_id);
    IntList face = new IntList();
    face.append(v_id); /* on remplace le sommet actuel par un des sommets de la future face */
    for(int i = 1; i < degree; ++i) {
      face.append(add_vertex(p.copy())); /* on ajoute les autres sommets qui composeront la face */
    }
    for(int f_id : faces)
      set_face(f_id, rebuild_face(v_id, get_face(f_id), neighbors, face));
    face.reverse();
    int id = add_face(face);
    flat_vertex_normals.add(vertex_normal);
    check_and_change_orientation();
    return id;
  }
  
  abstract void notif_modif_face(int f_id, IntList old_face);
  
  @Override
  int add_face(IntList face) {
    int f_id = data.add(face);
    notif_modif_face(f_id, null);
    return f_id;
  }
  
  @Override
  void set_face(int f_id, IntList face) {
    IntList old_face = data.faces.get(f_id);
    data.set(f_id, face);
    notif_modif_face(f_id, old_face);
  }
  
  abstract void notif_add_vertex(int v_id);
  
  @Override
  int add_vertex(PVector v) {
    int v_id = data.add(v);
    notif_add_vertex(v_id);
    return v_id;
  }
}
