/* ---------------------------------------------------------------------------- */
/* AdjacencyListSurface                                                         */
/* ---------------------------------------------------------------------------- */

/* Surface proposant la gestion des arêtes par des listes d'adjacence.         */
/* Hérite des fonctionnalités de base de AbstractSurface.                       */
/* L'intérêt principal de cette structure de données est que la liste           */
/* d'adjacence permet de récupérer la liste des arêtes et des faces sur        */
/* lesquelles on trouve un sommêt p en un temps linéaire en le voisinage de p.  */
/* L'inconvénient de cette structure de données est que vérifier qu'une arête  */
/* existe ou qu'un sommêt est sur une face donnée est aussi en temps linéaire   */
/* sur la taille du voisinage d'un sommêt. (ClassicSurface plus adaptée dans ce */
/* cas)                                                                         */

/* La classe Edge est une composante de la liste d'adjacence.                   */
/* Ici on admet qu'elle est définie pour un sommêt donné, on enregistre un      */
/* sommêt voisin (arête) et les faces sur lesquelles cette arête est présente */
class Edge {
  int vertex;
  IntList faces; /* doit être au plus 2 */
  
  Edge(int vertex) {
    this.vertex = vertex;
    faces = new IntList();
  }
  
  /* ajoute l'information d'une face seulement si elle n'est pas présente */
  void add_face_on(int face_id) {
    if(faces.hasValue(face_id))
      return;
    faces.append(face_id);
  }
}

class AdjacencyListSurface extends AbstractSurface implements Surface {
  /* Ici chaque ArrayList<Edge> correspond à une liste d'adjacence pour un sommêt donné. */
  /* adjacency est l'ensemble des listes d'adjacence du graphe. */
  ArrayList<ArrayList<Edge>> adjacency;
  
  /* sous-routines privées pour la construction de la structure de données */
  
  /* ajoute une arête et faces correspondante dans la liste d'adjacence */
  private void add_edge_and_face(int vertex, int other_vertex, int face_id) {
    for(Edge e : adjacency.get(vertex)) { /* recherche si l'arête (vertex, other_vertex) existe */
      if(e.vertex == other_vertex) {
        e.add_face_on(face_id); /* si la face n'a pas été traitée on ajoute l'information que cette arête est partagée par face_id */
        return;
      }
    }
    /* si l'arête (vertex, other_vertex) n'existe pas, on l'ajoute */
    Edge e = new Edge(other_vertex);
    e.add_face_on(face_id);
    adjacency.get(vertex).add(e);
  }
  
  /* change l'orientation d'une face si besoin (selon les orientations des faces précédentes)*/
  private void reverse_face_if_needed(IntList face) {
    for(int i = 0; i < face.size(); ++i) { /* Pour tous les sommêts de la face */
      for(Edge e : adjacency.get(face.get(i))) { /* On cherche dans ses voisins */
        if(e.vertex != face.get((i + 1) % face.size())) /* à obtenir l'autre face adjacente à une arête de cette face */
          continue;
        IntList other_face = get_face(e.faces.get(0));
        for(int j = 0; j < other_face.size(); ++j) { /* et corriger l'orientation de la face courante si nécessaire */
          if(other_face.get(j) == face.get(i) && face.get((i + 1) % face.size()) == other_face.get((j + 1) % other_face.size())) {
            face.reverse();
            return;
          }
        }
      }
    }
  }
  
  /* calcule la liste d'adjacence */
  private void build_list() {
    for(int f : faces_ids()) { /* on fabrique la liste à partir des informations données par les faces */
      IntList face = get_face(f);
      reverse_face_if_needed(face);
      for(int i = 0; i < face.size(); ++i) {
        int v1 = face.get(i);
        int v2 = face.get((i + 1) % face.size());
        /* on ajoute l'information de l'arête (v1, v2) présente sur la face f dans les listes d'adjacence de v1 et v2 */
        add_edge_and_face(v1, v2, f);
        add_edge_and_face(v2, v1, f);
      }
    }
    check_and_change_orientation();
  }
  
  AdjacencyListSurface(SurfaceData data) {
    super(data);
    adjacency = new ArrayList<ArrayList<Edge>>();
    for(int i : vertices_ids())
      adjacency.add(new ArrayList<Edge>());
    build_list();
  }
  
  int vertex_degree(int p) {
    return adjacency.get(p).size();
  }
  
  IntList neighbors(int p) {
    IntList res = new IntList();
    for(Edge e : adjacency.get(p))
      res.append(e.vertex);
    return res;
  }
  
  IntList faces_in(int p) {
    IntList res = new IntList();
    for(Edge e : adjacency.get(p))
      for(int q : e.faces)
        if(! res.hasValue(q))
          res.append(q);
    return res;
  }
  
  IntList orientedNeighbors(int p) {
    ArrayList<Edge> neighbors = adjacency.get(p);
    IntList res = new IntList();
    Edge current = neighbors.get(0); /* on sélectionne un premier voisin */
    res.append(current.vertex);
    while(res.size() < neighbors.size()) {
      neighbors_loop : for(Edge e : neighbors) { /* on cherche le voisin sur la face suivante */
        if(e.vertex == res.get(res.size() - 1))
          continue;
          
        for(int i = 0; i < e.faces.size(); ++i) {
          if(e.faces.get(i) == current.faces.get(0)) {
            res.append(e.vertex);
            current = new Edge(e.vertex); /* on construit le segment courrant de sorte que la première face soit la face de référence (orientation) */
            if(e.faces.size() > 1)
              current.add_face_on(e.faces.get((i + 1) & 1)); /* on récupère la face non étudiée (x & 1 <=> x % 2)*/
            else
              current.add_face_on(e.faces.get(0));
            break neighbors_loop;
          }
        }
      }
    }
    return res;
  }
  
  ArrayList<Edge> fullDataNeighbors(int p) {
    return adjacency.get(p);
  }
  
  void notif_modif_face(int f_id, IntList old_face) {
    if(old_face != null) {
      for(int qi : old_face) {
        for(Edge e : adjacency.get(qi))
          e.faces.removeValue(f_id);
        ArrayList<Edge> to_remove = new ArrayList<Edge>();
        for(Edge e : adjacency.get(qi)) {
          if(e.faces.size() == 0) {
            to_remove.add(e);
          }
        }
        for(Edge e : to_remove) {
          adjacency.get(qi).remove(e);
        }
      }
    }
    IntList face = get_face(f_id);
    
    for(int i = 0; i < face.size(); ++i) {
      add_edge_and_face(face.get((i + 1) % face.size()), face.get(i), f_id);
      add_edge_and_face(face.get((i + 1) % face.size()), face.get((i + 2) % face.size()), f_id);
    }
  }
  
  void notif_add_vertex(int v_id) {
    adjacency.add(new ArrayList<Edge>());
  }
}
