/* ---------------------------------------------------------------------------- */
/* GuiComponent                                                                 */
/* ---------------------------------------------------------------------------- */

/* Composants de l'interface graphique. Pour ajouter un composant à l'interface */
/* il suffit qu'il définisse les méthodes données par GuiComponent.             */
/* Implémentés pour le moments :                                                */
/* - GuiComponentOffsetProxy : composant qui permet un déplacement d'un         */
/* composant (permet de ne pas avoir à gérer les offsets dans chaque composant) */
/* - CheckBoxGuiComponent : boîte cliquable dont la valeur est un boolean.      */
/* - SelectorGuiComponent : permet la sélection d'un paramètre dans une liste.  */
/* - BarGuiComponent : ajuste la valeur flotante d'un paramètre.                */
/* À noter qu'il faut récupérer la valeur du paramètre par getValue() et qu'il  */
/* est possible de trouver le paramètre dans une liste générique de GuiComponent*/
/* en comparant le nom du paramètre cherché à getName(). (Ceci pourra être      */
/* remplacé par un VarHandle dans une version plus évoluée de Java.             */

interface GuiComponent {
  /* vérifie que les coordonnées sont dans le composant */
  boolean checkIn(float x, float y);
  /* fait les actions si actif */
  void doActions(float x, float y);
  /* fait des actions à l'activation */
  void getIn(float x, float y);
  /* fait des actions à la libération */
  void getOut(float x, float y);
  /* s'affiche à l'écran */
  void draw();
  /* récupère le nom du paramètre enregistré */
  String getName();
  /* récupèrela valeur du paramètre */
  Object getValue();
  /* donne la hauteur du composant */
  double getHeight();
}

abstract class AbstractGuiComponent implements GuiComponent {
  @Override
  void doActions(float x, float y) {
  }
  @Override
  void getIn(float x, float y) {
  }
  @Override
  void getOut(float x, float y) {
  }
}

class GuiComponentOffsetProxy implements GuiComponent {
  PVector offset;
  GuiComponent cmp;
  
  GuiComponentOffsetProxy(PVector offset, GuiComponent cmp) {
    this.offset = offset;
    this.cmp = cmp;
  }
  
  @Override
  boolean checkIn(float x, float y) {
    return cmp.checkIn(x - offset.x, y - offset.y);
  }
  
  @Override
  void doActions(float x, float y) {
    cmp.doActions(x - offset.x, y - offset.y);
  }
  
  @Override
  void getIn(float x, float y) {
    cmp.getIn(x - offset.x, y - offset.y);
  }
  
  @Override
  void getOut(float x, float y) {
    cmp.getOut(x - offset.x, y - offset.y);
  }
  
  @Override
  void draw() {
    pushMatrix();
    translate(offset.x, offset.y);
    cmp.draw();
    popMatrix();
  }
  
  @Override
  String getName() {
    return cmp.getName();
  }
  
  @Override
  Object getValue() {
    return cmp.getValue();
  }
  
  @Override
  double getHeight() {
    return cmp.getHeight();
  }
}

class CheckBoxGuiComponent extends AbstractGuiComponent implements GuiComponent {
  String name;
  boolean value;
  float size;
  
  CheckBoxGuiComponent(String name, float size, boolean value) {
    this.name = name;
    this.value = value;
    this.size = size;
  }
  
  @Override
  boolean checkIn(float x, float y) {
    return (new PVector(x, y)).mag() < size;
  }
  
  @Override
  void getIn(float x, float y) {
    if(checkIn(x, y))
      value = ! value;
  }
  
  @Override
  void draw() {
    stroke(250);
    if(value)
      fill(0x1d, 0x93, 0x36);
    else
      fill(0xa8, 0x1e, 0x1e);
    ellipse(0, 0, size, size);
    if(other_style)
      fill(0);
    else
      fill(250);
    stroke(250);
    text(name, (int)size, 5);
  }
  
  @Override
  String getName() {
    return name;
  }
  
  @Override
  Object getValue() {
    return value;
  }
  
  @Override
  double getHeight() {
    return size * 2;
  }
}

class SelectorGuiComponent <T> extends AbstractGuiComponent implements GuiComponent {
  class Option {
    String name;
    T value;
    
    Option(String name, T value) {
      this.name = name;
      this.value = value;
    }
  }
  String name;
  ArrayList<Option> options = new ArrayList<Option>();
  int current_id = 0;
  float size;
  
  SelectorGuiComponent(String name, float size) {
    this.name = name;
    this.size = size;
  }
  
  void add(String name, T option) {
    options.add(new Option(name, option));
  }
  
  @Override
  boolean checkIn(float x, float y) {
    return x > -size && x < size && y > 10 && y < 10 + 2 * size * options.size();
  }
  
  @Override
  void getIn(float x, float y) {
    if(! checkIn(x, y))
      return;
    current_id = (int)((y - 10) / (2 * size));
  }
  
  @Override
  void draw() {
    if(other_style)
      stroke(0);
    else
      stroke(250);
    text(name, -(int)(5 + size), 5);
    line(-size, 10, -size, 10 + 2 * size * options.size());
    line(-size, 10 + 2 * size * options.size(), -size + 50, 10 + 2 * size * options.size());
    for(int i = 0; i < options.size(); ++i) {
      stroke(250);
      if(i == current_id)
        fill(0xee, 0xbe, 0x4a);
      else
        fill(0x29, 0x29, 0x29);
      ellipse(0, 10 + (2 * i + 1) * size, size, size);
      if(other_style)
        fill(0);
      else
        fill(250);
      text(options.get(i).name, (int)(size), (int)(10 + (2 * i + 1) * size + 5));
    }
  }
  
  @Override
  String getName() {
    return name;
  }
  
  @Override
  T getValue() {
    if(current_id < 0 || current_id >= options.size())
      return null;
    return options.get(current_id).value;
  }
  
  @Override
  double getHeight() {
    return 2 * size * (1 + options.size());
  }
}

class BarGuiComponent extends AbstractGuiComponent implements GuiComponent {
  String name;
  float min_val;
  float max_val;
  float val;
  float b_height;
  float b_width;
  Function<Float, Float> non_linear;
  boolean isIn = false;
  
  BarGuiComponent(String name, float val, float min_val, float max_val, float b_height, float b_width, Function<Float, Float> non_linear) {
    this.name = name;
    this.val = val;
    this.min_val = min_val;
    this.max_val = max_val;
    this.b_height = b_height;
    this.b_width = b_width;
    this.non_linear = non_linear;
  }
  
  BarGuiComponent(String name, float val, float min_val, float max_val, float b_height, float b_width) {
    this(name, val, min_val, max_val, b_height, b_width, new Function<Float, Float>(){
      @Override
      Float apply(Float val) {
        return val;
      }
    });
    this.name = name;
    this.val = val;
    this.min_val = min_val;
    this.max_val = max_val;
    this.b_height = b_height;
    this.b_width = b_width;
  }
  
  @Override
  boolean checkIn(float x, float y) {
    return x > -b_height && x < b_width + b_height && y > -b_height && y < b_height;
  }
  
  @Override
  void getIn(float x, float y) {
    isIn = true;
  }
  
  @Override
  void getOut(float x, float y) {
    isIn = false;
  }
  
  @Override
  void doActions(float x, float y) {
    if(x < 0) {
      val = min_val;
      return;
    }
    else if(x > b_width) {
      val = max_val;
      return;
    }
    val = (x / b_width) * (max_val - min_val) + min_val;
  }
  
  @Override
  void draw() {
    if(other_style)
      stroke(0);
    else
      stroke(250);
    line(0, 0, -1, b_width, 0, -1);
    if(isIn)
      fill(0xee, 0xbe, 0x4a, 216);
    else {
      if(other_style)
        fill(0, 0, 0, 216);
      else
        fill(0x29, 0x29, 0x29, 216);
    }
    float pos = ((val - min_val) / (max_val - min_val)) * b_width;
    ellipse(pos, 0, b_height, b_height);
    if(other_style)
      fill(0);
    else
      fill(250);
    text(name + " : " + getValue(), (int)(b_width + b_height), 5);
  }
  
  @Override
  String getName() {
    return name;
  }
  
  @Override
  Float getValue() {
    return non_linear.apply(val);
  }
  
  @Override
  double getHeight() {
    return 2 * b_height;
  }
}
