/* ---------------------------------------------------------------------------- */
/* MeanCurvature                                                                */
/* ---------------------------------------------------------------------------- */

/* Le flot de courbure moyenne est le flot qui diminue au mieux l'aire d'une    */
/* surface.                                                                     */

/* ---------------------------------------------------------------------------- */
/* VerticesMeanCurvature                                                        */
/* ---------------------------------------------------------------------------- */

/* Ici est implémenté la méthode qui détermine la courbure moyenne par la       */
/* formule des cotangentes. Pour tout sommêt p, le flot de courbure moyen par   */
/* les sommêts est donné par la somme pour qi voisin de p :                     */
/* (1/2) (cotan((Qi,Q{i-1},P)) + cotan((Qi,Q{i+1},P))) (Qi,P)                   */

class VerticesMeanCurvature implements CurvatureEngine {
  boolean probabilistic = false;
  
  VerticesMeanCurvature() {
  }
  
  VerticesMeanCurvature(boolean probabilistic) {
    this.probabilistic = probabilistic;
  }
  
  PVectorList build(Surface surface) {
    PVectorList res = new PVectorList();
    for(int p : surface.vertices_ids()) {
      PVector tmp = new PVector(0, 0, 0);
      PVector working = surface.get(p);
      ArrayList<PVector> neighbors = surface.vectors_from_ids(surface.orientedNeighbors(p));
      for(int i = 0; i < neighbors.size(); ++i) {
        PVector previous = neighbors.get((i - 1 + neighbors.size()) % neighbors.size()); /* alias q_{i - 1}*/
        PVector current = neighbors.get(i); /* alias q_i*/
        PVector next = neighbors.get((i + 1) % neighbors.size()); /* alias q_{i + 1}*/
        float coeff = cotan(PVector.angleBetween(PVector.sub(current, previous), PVector.sub(working, previous)));
        coeff += cotan(PVector.angleBetween(PVector.sub(current, next), PVector.sub(working, next)));
        tmp.add(PVector.mult(PVector.sub(working, current), coeff / 2));
      }
      if(probabilistic)
        tmp.div(neighbors.size());
      res.add(tmp);
    }
    return res;
  }
}

/* ---------------------------------------------------------------------------- */
/* EdgesMeanCurvature                                                           */
/* ---------------------------------------------------------------------------- */

/* Ici, on propose une version de la mean curvature qui se base sur les arêtes.*/
/* Notons qu'elle permet de résoudre certains problème de la courbure moyenne   */
/* sur les arêtes lorsque les faces deviennent très petites ou les angles sur  */
/* les faces deviennent de même trop petits pour que le calcul ne subisse pas   */
/* d'instabilité numérique dû aux valeurs flotantes.                            */
/* La forumule implémentée ici est la somme pour e une arête et theta_e l'angle*/
/* dihédral sur l'arête e (angle entre les normales des faces en e) des        */
/* -(|e|/2)sin(theta_e/2)N(e) où N(e) est la normale en e.                      */

import java.util.function.Function;

PVector get_He(int v_id, Edge e, Surface surface) {
  PVector N_f1 = surface.face_normal(e.faces.get(0));
  PVector N_f2 = surface.face_normal(e.faces.get((e.faces.size() > 1) ? 1 : 0));
  IntList f1 = surface.get_face(e.faces.get(0));
  if(edge_oriented_on_face(v_id, e.vertex, f1)) {
    PVector tmp = N_f1;
    N_f1 = N_f2;
    N_f2 = tmp;
  }
  PVector Ne = PVector.sub(N_f2, N_f1).cross(surface.get(v_id, e.vertex));
  Ne.div(-2);
  return Ne;
}

class EdgesMeanCurvature extends EdgeCurvatureEngine implements CurvatureEngine {
  private Function<PVector, Float> anisotropic_function;
  
  EdgesMeanCurvature(Function<PVector, Float> anisotropic_function) { /* Laisse la possibilité de définir une fonction de poids sur les arêtes */
    this.anisotropic_function = anisotropic_function;
  }
  
  EdgesMeanCurvature() {
    this(new Function<PVector, Float>() { /* donne le poids 1 pour toute arête : courbure moyenne sur les arêtes classique */
      @Override
      Float apply(PVector input) {
        return 1.;
      }
    });
  }
  
  @Override
  PVector[][] build_edge_vectors(Surface surface) {
    PVector [][] res = new PVector[surface.nV()][surface.nV()];
    
    for(int v_id : surface.vertices_ids()) {
      for(Edge e : surface.fullDataNeighbors(v_id)) {
        PVector Ne = get_He(v_id, e, surface);
        Ne.mult(anisotropic_function.apply(Ne));
        res[v_id][e.vertex] = Ne;
      }
    }
    return res;
  }
}

/* ---------------------------------------------------------------------------- */
/* FacesMeanCurvature                                                           */
/* ---------------------------------------------------------------------------- */

/* Ici est une autre implémentation de la courbure moyenne définie par les faces*/
/* On définit le flot du courbure sur star(P) où les e_i sont les arêtes       */
/* ordonnées (P,Qi)et où Qi et Q{i+1} sont sur la face f_i. De même on définit  */
/* les theta_{e_i} les angles dihédraux sur les arêtes e_i et N(f_i) les       */
/* normales sur les faces f_i. Le calcul du flot moyen sur les sommêts est donné*/
/* par la somme des                                                             */
/* (-1/4)(|e_i|tan(theta_{e_i}) + |e_{i+1}|tan(theta_{e_{i+1}})N(f_i)           */

class FacesMeanCurvature implements CurvatureEngine {
  @Override
  PVectorList build(Surface surface) {
    PVectorList res = new PVectorList();
    ArrayList<PVector> face_normals = new ArrayList<PVector>();
    for(int f_id : surface.faces_ids()) {
      face_normals.add(surface.face_normal(f_id));
    }
    for(int v_id : surface.vertices_ids()) {
      PVector working = new PVector(0, 0, 0);
      HashMap<Integer, FloatList> coeffs = new HashMap<Integer, FloatList>(); /* values of the (|ei|/2)tan(theta_ei/2)*/
      
      for(Edge e : surface.fullDataNeighbors(v_id)) {
        PVector N_f1 = surface.face_normal(e.faces.get(0));
        PVector N_f2 = surface.face_normal(e.faces.get((e.faces.size() > 1) ? 1 : 0));
        PVector N_sum = PVector.add(N_f1, N_f2);
        IntList f1 = surface.get_face(e.faces.get(0));
        if(edge_oriented_on_face(v_id, e.vertex, f1)) {
          PVector tmp = N_f1;
          N_f1 = N_f2;
          N_f2 = tmp;
        }
        float c = PVector.sub(N_f2, N_f1).cross(surface.get(v_id, e.vertex)).dot(N_sum) / N_sum.dot(N_sum);
        /* PVector.sub(N_f1, N_f2).cross(surface.get(v_id, e.vertex)) est en fait |ei|sin(theta_ei/2)N(ei) and N(ei) = N_sum / (2 * cos(theta_ei/2)) */
        /* Donc correspond à (|ei|/2)tan(theta_ei/2)N_sum, par suite la bilinéarité du produit scalaire nous permet de récupérer le coefficient (|ei|/2)tan(theta_ei/2) */
        for(int f_id : e.faces) {
          coeffs.putIfAbsent(f_id, new FloatList());
          coeffs.get(f_id).append(c);
        }
        
      }
        
      for(int f_id : surface.faces_in(v_id)) {
        float coeff = 0;
        for(float c : coeffs.get(f_id))
          coeff += c;
        coeff /= -2;
        working.add(PVector.mult(surface.face_normal(f_id), coeff));
      }
      working.normalize();
      res.add(working);
    }
    return res;
  }
}

/* ---------------------------------------------------------------------------- */
/* HildebrandtPolthierAnisotropicMeanCurvature                                  */
/* ---------------------------------------------------------------------------- */

/* Ici est proposé une version du flot de courbure moyen sur les arêtes        */
/* renormalisé à l'aide d'une fonction de poids w proposée par Hildebrandt et   */
/* Polthier. w est définie sur les arêtes e pour paramètres réels lambda et r :*/
/* w(e) = { 1 si |e| < lambda                                                   */
/*          lambda^2 / (r(|e| - lambda)^2 + lambda^2) sinon }                   */

class HildebrandtPolthierAnisotropicFunction implements Function<PVector, Float> {
  float lambda;
  float r;
  
  HildebrandtPolthierAnisotropicFunction(float lambda, float r) {
    this.lambda = lambda;
    this.r = r;
  }
  
  @Override
  Float apply(PVector input) {
    float x = input.mag();
    if(x <= lambda)
      return 1.;
    return (lambda * lambda) / (r * pow(x - lambda, 2) + lambda * lambda);
  }
}

EdgesMeanCurvature HildebrandtPolthierEdgesMeanCurvature(float lambda, float r) {
  return new EdgesMeanCurvature(new HildebrandtPolthierAnisotropicFunction(lambda, r));
}
