/* ---------------------------------------------------------------------------- */
/* PMatrix                                                                      */
/* ---------------------------------------------------------------------------- */

/* Un outil pour la manipulation de matrices 3x3. Sa principale utilisation est */
/* le calcul des vecteurs propres depuis l'opérateur de forme.                  */
/* Pour obtenir les vecteurs propres la méthode utilisée est la décompositon QR */
/* et l'utilisation de la méthode itérative du QR algorithm.                    */

class PMatrix {
  PVectorList vectors;
  
  PMatrix(PVectorList vectors) {
    this.vectors = vectors;
  }
  
  PMatrix(PVector v1, PVector v2, PVector v3) {
    vectors = new PVectorList();
    vectors.add(v1);
    vectors.add(v2);
    vectors.add(v3);
  }
  
  PMatrix() {
    vectors = new PVectorList();
    vectors.add(new PVector(1, 0, 0));
    vectors.add(new PVector(0, 1, 0));
    vectors.add(new PVector(0, 0, 1));
  }
  
  PVector mult(PVector v) {
    return new PVector(
      vectors.get(0).dot(v),
      vectors.get(1).dot(v),
      vectors.get(2).dot(v)
    );
  }
  
  PMatrix copy() {
    return new PMatrix(vectors.copy());
  }
  
  PMatrix mult(float s) {
    for(PVector v : vectors)
      v.mult(s);
    return this;
  }
  
  PMatrix transpose() {
    return new PMatrix(
      new PVector(get(0).x, get(1).x, get(2).x),
      new PVector(get(0).y, get(1).y, get(2).y),
      new PVector(get(0).z, get(1).z, get(2).z)
    );
  }
  
  PMatrix mult(PMatrix other) {
    PMatrix t = other.transpose();
    return new PMatrix(
      new PVector(get(0).dot(t.get(0)), get(0).dot(t.get(1)), get(0).dot(t.get(2))),
      new PVector(get(1).dot(t.get(0)), get(1).dot(t.get(1)), get(1).dot(t.get(2))),
      new PVector(get(2).dot(t.get(0)), get(2).dot(t.get(1)), get(2).dot(t.get(2)))
    );
  }
  
  PMatrix add(PMatrix other) {
    for(int i = 0; i < vectors.size(); ++i)
      vectors.get(i).add(other.vectors.get(i));
    return this;
  }
  
  float trace() {
    return get(0).x + get(1).y + get(2).z;
  }
  
  PVector get(int i) {
    return vectors.get(i);
  }
  
  PVector diagonal() {
    return new PVector(get(0).x, get(1).y, get(2).z);
  }
  
  float det() {
    return get(0).dot(get(1).cross(get(2)));
  }
  
  @Override
  String toString() {
    return "[" + get(0) + "\n " + get(1) + "\n " + get(2) + "]";
  }
  
  float dist(PMatrix other) {
    PVectorList l = vectors.sub(other.vectors);
    return l.dot(l);
  }
  
  PMatrix inverse() {
    PMatrix Q = QR_matrix(this);
    PMatrix R = Q.transpose().mult(this);
    PMatrix inv_R = (new PMatrix(
      solve(new PVector(1, 0, 0), R),
      solve(new PVector(0, 1, 0), R),
      solve(new PVector(0, 0, 1), R)
    )).transpose();
    return inv_R.mult(Q.transpose());
  }
}

PMatrix tensor_product(PVector v1, PVector v2) {
  PVectorList vectors = new PVectorList();
  vectors.add(PVector.mult(v2, v1.x));
  vectors.add(PVector.mult(v2, v1.y));
  vectors.add(PVector.mult(v2, v1.z));
  return new PMatrix(vectors);
}

/* résout un système d'équations linéaires pour une matrice triangulaire */
/* Note : utile depuis la décomposition QR (implémenter dans le cas général si besoin) */
/* À noter que cette méthode est utilisée pour inverser la matrice : utiliser PMatrix.inverse() dans le cas général */
PVector solve(PVector v, PMatrix triangulaire) {
  PVector res = new PVector(0, 0, 0);
  try {
    res.z = v.z / triangulaire.get(2).z;
    res.y = (v.y - res.z * triangulaire.get(1).z) / triangulaire.get(1).y;
    res.x = (v.x - res.y * triangulaire.get(0).y - res.z * triangulaire.get(0).z) / triangulaire.get(0).x;
  } catch(Exception e) {
    throw new IllegalArgumentException("The matrix needs to be invertible to be inversed");
  }
  return res;
}

/* QR décomposition : retourne Q */
/* notons que mat = Q * R avec Q orthogonale (Q * tQ = I) et R triangulaire supérieure */
PMatrix QR_matrix(PMatrix mat) {
  float r, c, s;
  c = mat.get(0).x;
  s = -mat.get(1).x;
  r = sqrt(c * c + s * s);
  c /= r;
  s /= r;
  PMatrix G1 = new PMatrix(
    new PVector(c, -s, 0),
    new PVector(s, c, 0),
    new PVector(0, 0, 1)
  );
  PMatrix A2 = G1.mult(mat); /* élimine ligne 2 colonne 1 */
  c = A2.get(0).x;
  s = -A2.get(2).x;
  r = sqrt(c * c + s * s);
  c /= r;
  s /= r;
  PMatrix G2 = new PMatrix( /* élimine ligne 3 colonne 1 */
    new PVector(c, 0, -s),
    new PVector(0, 1, 0),
    new PVector(s, 0, c)
  );
  PMatrix A3 = G2.mult(A2);
  c = A3.get(1).y;
  s = -A3.get(2).y;
  r = sqrt(c * c + s * s);
  c /= r;
  s /= r;
  PMatrix G3 = new PMatrix( /* élimine ligne 3 colonne 2 */
    new PVector(1, 0, 0),
    new PVector(0, c, -s),
    new PVector(0, s, c)
  );
  PMatrix Q = G3.mult(G2).mult(G1).transpose();
  /*PMatrix R = Q.transpose().mult(mat);*/
  /*println("R");
  println(R);*/
  return Q;
}

FloatList PVector_to_list(PVector v) {
  FloatList res = new FloatList();
  res.append(v.x);
  res.append(v.y);
  res.append(v.z);
  return res;
}

/* implémentation du QR algorithm pour la recherche de vecteurs propres */
PVectorList eigenVectors(PMatrix mat) {
  PMatrix Q = new PMatrix();
  PMatrix A = mat;
  PMatrix prev = new PMatrix();
  for(int i = 0; A.dist(prev) > 1 && i < 100; ++i) {
    PMatrix Q1 = QR_matrix(A);
    PMatrix R1 = Q1.transpose().mult(A);
    Q = Q.mult(Q1);
    prev = A;
    A = R1.mult(Q1);
  }
  return Q.transpose().vectors;
}
