/* ---------------------------------------------------------------------------- */
/* PVectorList                                                                  */
/* ---------------------------------------------------------------------------- */

/* La PVectorList est une liste de PVector avec des fonctionnalités             */
/* supplémentaire par rapport à une liste classique : elle peut être manipulée  */
/* comme un champs de vecteurs.                                                 */
/* Utile pour la création et manipulation d'un flot.                            */
/* Elle est itérable : autorise les foreach par exemple (for(PVector v : liste))*/
/* Une méthode correct est proposée pour corriger les vecteurs de norme infinie */
/* ou de norme NaN obtenus par instabilité numérique ou par exemple en cas de   */
/* calcul sur des vecteurs identiques.                                          */

import java.util.Iterator;

/* Fabrique un PVectorList nulle d'une taille donnée. */
PVectorList PVectorList_zero(int size) {
  PVectorList res = new PVectorList();
  for(int i = 0; i < size; ++i)
    res.add(new PVector(0, 0, 0));
  return res;
}

class PVectorList implements Iterable<PVector>, VectorField<PVectorList> {
  
  ArrayList<PVector> list;
  
  PVectorList() {
    list = new ArrayList<PVector>();
  }
  
  PVectorList(int size) {
    list = new ArrayList<PVector>(size);
  }
  
  PVectorList(ArrayList<PVector> other) {
    this();
    for(PVector v : other)
      list.add(v);
  }
  
  PVectorList vectors_from_ids(IntList ids) {
    PVectorList res = new PVectorList();
    for(int id : ids)
      res.add(get(id));
    return res;
  }
  
  Iterator<PVector> iterator() {
    return list.iterator();
  }
  
  PVector get(int id) {
    return list.get(id);
  }
  
  void set(int id, PVector obj) {
    list.set(id, obj);
  }
  
  int size() {
    return list.size();
  }
  
  PVectorList copy() {
    PVectorList res = new PVectorList();
    for(PVector v : this)
      res.add(v.copy());
    return res;
  }
  
  void add(PVector v) {
    list.add(v);
  }
  
  ArrayList<PVector> get_as_vector() {
    return list;
  }
  
  float dot(PVectorList other) {
    float sum = 0;
    for(int i = 0; i < min(size(), other.size()); ++i)
      sum += PVector.dot(get(i), other.get(i));
    return sum;
  }
  
  float norm() {
    return sqrt(dot(this));
  }
  
  PVectorList mult(float scalar) {
    PVectorList res = new PVectorList();
    for(PVector v : this)
      res.add(PVector.mult(v, scalar));
    return res;
  }
  
  PVectorList add(PVectorList other) {
      PVectorList res = new PVectorList();
      for(int i = 0; i < min(size(), other.size()); ++i)
        res.add(PVector.add(get(i), other.get(i)));
      return res;
  }
  
  PVectorList sub(PVectorList other) {
      PVectorList res = new PVectorList();
      for(int i = 0; i < min(size(), other.size()); ++i)
        res.add(PVector.sub(get(i), other.get(i)));
      return res;
  }
  
  PVectorList projection(PVectorList... basis) {
    if(basis.length == 0)
      throw new IllegalArgumentException("Please project over something");
    PVectorList res = PVectorList_zero(size());
    for(PVectorList l : basis)
      res = res.add(l.mult(dot(l) / l.norm()));
    return res;
  }
  
  void correct() { /* corrects the vector value in case of numerical instability and prevents from some application crash */
    for(int i = 0; i < size(); ++i)
      if(get(i).mag() > Float.MAX_VALUE || Float.isNaN(get(i).mag()))
          set(i, new PVector(0, 0, 0));
  }
  
  PVector barycenter() {
    PVector res = new PVector(0, 0, 0);
    for(PVector v : this)
      res.add(v);
    res.div(size());
    return res;
  }
  
  @Override
  String toString() {
    return list.toString();
  }
  
  void clean_doubles() {
    clean_doubles(1e-6);
  }
  
  void clean_doubles(float error_dist) {
    for(int i = 0; i < size(); ++i) {
      for(int j = i + 1; j < size(); ++j) {
        if(PVector.sub(get(i), get(j)).mag() < error_dist) {
          list.remove(j);
          --j;
        }
      }
    }
  }
}
