/* ---------------------------------------------------------------------------- */
/* PrincipalCurvatures                                                          */
/* ---------------------------------------------------------------------------- */

/* Ici est proposé un outil de calculation des courbures principales.*/

abstract class PrincipalCurvaturesEngine implements CurvatureEngine {
  PVectorList Kmax;
  PVectorList Kmin;
  PVectorList gradient;
  
  /* Méthode à redéfinir qui retourne le vecteur du flot fabriqué depuis les courbures principales */
  /* notons que le gradient est calculé et accessible dans cette méthode depuis super.gradient */
  abstract PVector get_from_principals(Surface surface, int v_id, PVector k1, PVector k2);
  
  @Override
  PVectorList build(Surface surface) {
    PVectorList res = new PVectorList();
    gradient = partialGradient(surface);
    Kmax = new PVectorList();
    Kmin = new PVectorList();
    
    for(int v_id : surface.vertices_ids()) {
      /* on fabrique l'opérateur de forme en un sommet */
      PMatrix Sp = vertex_shape_operator(surface, v_id, gradient.get(v_id));
      
      /* On résout le calcul de vecteurs propres et on sélectionne ceux associés aux deux plus grandes valeurs propres en valeur absoule */
      solve_principals(Sp, Kmax, Kmin);
    }
    /* On régularise le signe des vecteurs de Kmin et Kmax de sorte d'avoir une régularité sur un maximum de triangles */
    regularize_sign(surface);
    for(int v_id : surface.vertices_ids()) {
      res.add(get_from_principals(surface, v_id, Kmax.get(v_id), Kmin.get(v_id)));
    }
    return res;
  }
  
  private void regularize_sign(Surface surface) {
    for(int v_id : surface.vertices_ids()) {
      for(int qi : surface.neighbors(v_id)) {
        if(qi > v_id)
          continue;
        if(Kmax.get(v_id).dot(Kmax.get(qi)) < 0) {
          Kmax.get(v_id).mult(-1);
        }
        if(Kmin.get(v_id).dot(Kmin.get(qi)) < 0) {
          Kmin.get(v_id).mult(-1);
        }
      }
    }
  }
}

PMatrix vertex_shape_operator(Surface surface, int v_id, PVector Np) {
  PMatrix Sp = new PMatrix(new PVector(0, 0, 0), new PVector(0, 0, 0), new PVector(0, 0, 0));
  for(Edge e : surface.fullDataNeighbors(v_id)) {
    PVector He = get_He(v_id, e, surface);
    PVector Ne = PVector.div(He, He.mag());
    /* On fabrique He et on en déduit Ne pour fabriquer l'opérateur de forme */
    /* Ici, il est plus simple informatiquement de le gérer sous forme de matrice obtenue par produit tensoriel */
    PMatrix Se = shape_operator(He, surface.get(v_id, e.vertex));
    /*println("He mag : " + He.mag());
    println("Se trace : " + Se.trace());*/
    /* On somme ensuite les opérateurs de forme définit sur chaque arête pour obtenir l'opérateur en v_id */
    Sp.add(Se.mult(Np.dot(Ne) / 2));
  }
  return Sp;
}

void solve_principals(PMatrix Sp, PVectorList Kmax, PVectorList Kmin) {
  /* On récupère les valeurs propres de l'opérateur de forme courant */
  PVectorList eigen_vectors = eigenVectors(Sp);
  PVector k1 = null, k2 = null;
  
  float k1val = 0, k2val = 0;
  float val;
  PVector tmp;
  for(PVector v : eigen_vectors) {
    tmp = Sp.mult(v);
    val = tmp.dot(tmp) / v.dot(v);
    /* on fait en sorte que k1 soit kmax et k2 kmin, */
    /* le vecteur propre de plus petite valeur propre en valeur absolue ne sera pas sauvegardé */
    /* conditions pour trier k1 et k2 sous les critères donnés plus haut : */
    if(k1 == null) { 
      k1 = v; k1val = val;
    } else if(k2 == null) {
      if(val > k1val) {
        k2 = k1; k2val = k1val;
        k1 = v; k1val = val;
      } else {
        k2 = v; k2val = val;
      }
    } else {
      if(val > k1val) {
        k1 = v; k1val = val;
      } else if(val > k2val) {
        k2 = v; k2val = val;
      }
    }
  }
  k1.normalize();
  k2.normalize();
  /* On ajoute les vecteurs calculés dans les listes Kmax et Kmin pour réutilisation ensuite */
  Kmax.add(k1);
  Kmin.add(k2);
}

PVector vertex_normal(Surface surface, int v_id, PVectorList gradient) {
  PVectorList Kdir = new PVectorList();
  PMatrix Sp = vertex_shape_operator(surface, v_id, gradient.get(v_id));
  solve_principals(Sp, Kdir, Kdir);
  PVector normal = Kdir.get(0).cross(Kdir.get(1));
  normal.normalize();
  if(normal.dot(gradient.get(v_id)) < 0)
    normal.mult(-1);
  return normal;
}

PMatrix shape_operator(PVector He, PVector e) {
  PVector te = e.cross(He);
  te.normalize();
  return tensor_product(te, te).mult(He.mag());
}

class GaussianCurvature extends PrincipalNormalCurvature implements CurvatureEngine {
  @Override
  float renormalization_coefficient(Surface surface, int v_id, PVector k1, PVector k2) {
    //return surface.star_area(v_id) / 3;
    /* renormalisation du flot pour obtenir une courbure gaussienne */
    return surface.star_area(v_id) / ((TWO_PI - surface.defaut_angulaire(v_id)) * 3);
  }
}

class PrincipalNormalCurvature extends PrincipalCurvaturesEngine implements CurvatureEngine {
  /* Renormalisation par défaut du flot de courbure normal, à redéfinir pour d'autres courbures et garder le coeur du code */
  float renormalization_coefficient(Surface surface, int v_id, PVector k1, PVector k2) {
    return 1;
  }
  
  PVector get_from_principals(Surface surface, int v_id, PVector k1, PVector k2) { /* Test */
    PVector res = k1.cross(k2);
    res.normalize();
    /* on ajuste le signe du vecteur de sorte qu'il soit dans la même direction que la pointe de star(v_id) */
    res = correct_normal(surface, res, v_id);
    /* On applique une renormalisation au vecteur qui est redéfinissable par héritage */
    res.mult(renormalization_coefficient(surface, v_id, k1, k2));
    return res;
  }
}
