/* ---------------------------------------------------------------------------- */
/* Surface                                                                      */
/* ---------------------------------------------------------------------------- */

/* Interface permettant de définir les fonctionnalités attendues pour qu'une    */
/* classe puisse être gérée comme une surface.                                  */
/* Pour ne pas avoir a écrire les fonctionnalités de base d'une surface,        */
/* La classe AbstractSurface est mise à disposition pour héritage de ses        */
/* fonctionnalités.                                                             */

interface Surface {
  /* Récupération d'une référence sur les données de la surface : positions, faces */
  SurfaceData get_data();
  
  /* Affichage de la surface à l'écran */
  void draw();
  
  /* Affichage de la surface à l'écran avec colorisation relative à une surface de référence */
  void draw_from_ground_truth(SurfaceData ground_truth);
  void draw_from_ground_truth_face_error(SurfaceData ground_truth);
  
  /* Affichage des vecteurs du flot sur la surface */
  void draw_flow(ArrayList<PVector> vector_field, float tau);
  /* Affichage des vecteurs du flot sur la surface pour une couleur donnée */
  void draw_flow(ArrayList<PVector> vector_field, float tau, int r, int g, int b);
  
  void draw_faces_normals(float size);
  
  /* obtenir une position d'un sommêt */
  PVector get(int vertex_id); 
  /* obtenir le vecteur qui définit l'arête d'un sommet à un autre */
  PVector get(int first, int second); 
  /* obtenir une face par son id */
  IntList get_face(int f_id);
  
  /* Itérateurs sur les composantes de la surface */
  /* à utiliser dans une boucle foreach : for(Type element : iterator()) */
  /* À noter que l'utilisation de ces fonctions peut améliorer la lisibilité et sont stables en cas de modification de la manière de gérer faces et sommêts */
  Iterable<PVector> vertices(); /* itérable sur les sommêts */
  Iterable<IntList> faces(); /* itérable sur les faces */
  Iterable<Integer> vertices_ids(); /* itérable sur les ids des sommêts : for(int i : [...]_ids()) <=> for(int i = 0; i < [...].size(); ++i) */
  Iterable<Integer> faces_ids(); /* itérable sur les ids des faces */
  
  int nV(); /* retourne le nombre de sommêts */
  int nF(); /* retourne le nombre de faces */
  
  /* changement d'échelle de la surface */
  void scale(float scalar); 
  /* déplacement de la surface */
  void translate(PVector direction); 
  
  /* donne le nombre de voisins à un sommêt donnée */
  int vertex_degree(int p);
  /* donne la liste des voisins d'un sommêt */
  IntList neighbors(int p); 
  /* donne toutes les faces sur lesquelles on trouve un sommêt */
  IntList faces_in(int p); 
  /* donne la liste des voisins d'un sommêt de sorte que chaque couple de voisins successifs soient sur la même face */
  IntList orientedNeighbors(int p); 
  /* Donne la normale de la face f (à noter que la normale dépend de l'orentation de la face) */
  PVector face_normal(int f);
  /* Fabrique la liste de toutes les normales des faces */
  PVectorList faces_normal();
  /* Donne la liste ordonnée des voisins en ajoutant l'information des faces où se trouvent chaque arête */
  /* Optimal pour la liste d'adjacence et intéressant dans certains calculs de courbures (par arêtes ou faces) */
  ArrayList<Edge> fullDataNeighbors(int p);
  
  /* permet de transformer une liste d'indices (sommêts) en la position de chaque sommêt */
  ArrayList<PVector> vectors_from_ids(IntList ids); 
  
  /* calcule le centre de masse de la surface */
  PVector barycenter(); 
  /* donne l'aire du voisinage d'un sommêt */
  float star_area(int p);
  /* donne l'aire non signée sur la surface */
  float area();
  /* donne le volume non signé de la surface */
  float volume();
  /* calcule le défaut angulaire au sommêt p */
  float defaut_angulaire(int p);
  /* change l'échelle de la surface de sorte que le plus long vecteur soit de norme max_size */
  void auto_scale(float max_size); 
  
  /* applique un flot à la surface renormalisé par un paramètre tau */
  void flow(ArrayList<PVector> vector_field, float tau);
  
  /* écrasement d'un sommet en une face */
  /* renvoie l'id de la face obtenue depuis le sommet */
  int flat(int p, PVector vertex_normal);
  /* ajoute une face à la surface */
  int add_face(IntList face);
  /* modifie une face donnée */
  void set_face(int f_id, IntList face);
  /* ajoute un sommet */
  int add_vertex(PVector v);
}
