/* ---------------------------------------------------------------------------- */
/* SurfaceData                                                                  */
/* ---------------------------------------------------------------------------- */

/* Donnée d'une surface : position des sommêts et faces (liste des indices des  */
/* sommêts pour construire une face)                                            */
/* L'intérêt de cette classe est de garder indépendants les données d'une       */
/* surface et la manière de gérer le graphe réprésentant la surface.            */

class SurfaceData {
  PVectorList vertices;
  ArrayList<IntList> faces;
  
  SurfaceData(int vertices_number, int faces_number) {
    vertices = new PVectorList(vertices_number);
    faces = new ArrayList<IntList>(faces_number);
  }
  
  SurfaceData(ArrayList<PVector> vertices, ArrayList<IntList> faces) {
    this.vertices = new PVectorList(vertices);
    this.faces = faces;
  }
  
  SurfaceData copy() {
    return new SurfaceData(vertices.copy().get_as_vector(), ArrayList_copy(faces));
  }
  
  int add(PVector p) {
    int id = vertices.size();
    vertices.add(p);
    return id;
  }
  
  int add(IntList face) {
    int id = faces.size();
    faces.add(face);
    return id;
  }
  
  void set(int v_id, PVector v) {
    vertices.set(v_id, v);
  }
  
  void set(int f_id, IntList f) {
    faces.set(f_id, f);
  }
}

/* Méthode de chargement d'une surface depuis un fichier .ply                   */
/* très inspirée de la méthode proposée par Pascal Romon (2018)                 */
/* À noter qu'une méthode factory est proposée ici pour éviter de construire    */
/* les données de la classe dans son constructeur (fragile et instable en java) */
/* peut lancer une IllegalArgumentException si le type de fichier est invalide  */
/* La SurfaceData obtenue peut être passée à un constructeur de surface pour    */
/* être gérée par une structure de données plus évoluée.                        */
SurfaceData loadPLY(String path) {
  String[] lines = loadStrings(path);
  
  ArrayList<PVector> vertices = new ArrayList<PVector>();
  ArrayList<IntList> faces = new ArrayList<IntList>();
  
  if(! lines[0].equals("ply"))
    throw new IllegalArgumentException("Given file is not a PLY file");
    
  int nF = 0;
  int nV = 0;
  
  int i = 0;
  String[] keywords;
  do { /* lecture des champs de l'entête du fichier */
    keywords = split(lines[i], ' ');
    if(keywords[0].equals("element")) {
      switch(keywords[1]) {
      case "vertex": /* nombre de sommêts */
        nV = int(keywords[2]);
        break;
      case "face": /* nombre de faces */
        nF = int(keywords[2]);
        break;
      }
    }
    ++i;
  } while(! keywords[0].equals("end_header"));
  
  for(int j = 0; j < nV; ++j, ++i) { /* lecture des positions des sommêts */
    keywords = split(lines[i], ' ');
    vertices.add(new PVector(float(keywords[0]), float(keywords[1]), float(keywords[2])));
  }
  
  for(int j = 0; j < nF; ++j, ++i) { /* lecture des indices définissant les faces */
    keywords = split(lines[i], ' ');
    int degree = int(keywords[0]);
    IntList current = new IntList();
    for(int k = 0; k < degree; ++k)
      current.append(int(keywords[k + 1]));
    faces.add(current);
  }
  
  println("Chargement de " + path + " réussi (sommets : " + nV + ", faces : " + nF + ")");
  
  return new SurfaceData(vertices, faces); /* construction des données */
}
