/* ---------------------------------------------------------------------------- */
/* SurfaceManager                                                               */
/* ---------------------------------------------------------------------------- */

/* Permet de faire le lien entre l'interface graphique et la manipulation des   */
/* surfaces.                                                                    */

/* SurfaceInfo est une sous-composante de SurfaceManager au cas où le code      */
/* évoluerait pour manipuler plusieurs surfaces en même temps.                  */

class SurfaceInfo {
  Surface surface;
  Surface ground_truth = null;
  SurfaceData previous = null;
  PVectorList flow = null;
  
  boolean do_curvature_on_ground_truth = false;
  boolean compare_to_previous = true;
  
  SurfaceInfo(Surface surface) {
    this.surface = surface;
    ground_truth = null;
    previous = null;
    flow = null;
  }
  
  void addGroundTruth(Surface ground_truth) {
    this.ground_truth = ground_truth;
  }
  
  /* centre la surface */
  void center() {
    this.surface.translate(PVector.mult(this.surface.barycenter(), -1));
    if(ground_truth != null)
      this.ground_truth.translate(PVector.mult(this.ground_truth.barycenter(), -1));
  }
  
  void prepare_flow(CurvatureEngine curvature) {
    flow = curvature.build(surface); /* On fabrique le flot de surface d'après la méthode donnée par curvature */
    flow.correct(); /* corrige les principales erreurs dues à l'instabilité numérique des flotants */
  }
  
  void apply_flow(SurfaceManager parent, CurvatureEngine curvature, float tau) {
    if(flow == null)
      prepare_flow(curvature);
    previous = surface.get_data().copy();
    
    surface.flow(flow.get_as_vector(), (minimize ? -1 : 1) * tau / flow.norm()); /* on déforme la surface par le flot calculé */
    ++parent.deform_count;
    
    if(ground_truth != null && do_curvature_on_ground_truth) {
      flow = curvature.build(ground_truth);
      flow.correct();
      ground_truth.flow(flow.get_as_vector(), (minimize ? -1 : 1) * tau / flow.norm());
    }
    prepare_flow(curvature);
  }
  
  /* mise à l'échelle à posteriori */
  void auto_scale() {
    surface.auto_scale(auto_scale_size);
    if(ground_truth != null)
      ground_truth.auto_scale(auto_scale_size);
  }
  
  void scale(float scalar) {
    surface.scale(scalar);
    if(ground_truth != null)
      ground_truth.scale(scalar);
  }
  
  void draw(SurfaceManager parent) {
    if(parent.do_color && ground_truth != null) { /* affiche la surface et les couleurs représentatives de l'erreur à la surface de référence */
      surface.draw_from_ground_truth(ground_truth.get_data());
    } else if (parent.do_color && compare_to_previous && previous != null) { /* affiche en comparant la surface déformée à la surface précédente */
      surface.draw_from_ground_truth(previous);
    } else {
      surface.draw(); /* cas om on affiche la surface sans couleurs */
    }
    if(parent.draw_vector_field && flow != null) { /* affiche de flot de courbure calculé */
      surface.draw_flow(flow.get_as_vector(), (minimize ? 1 : -1) * 10 * tau / flow.norm());
    }
    if(parent.draw_principals_flow && parent.curvature instanceof PrincipalCurvaturesEngine) { /* affiche le flot de courbure principale s'il est calculé */
        PrincipalCurvaturesEngine c = (PrincipalCurvaturesEngine)parent.curvature;
        surface.draw_flow(c.Kmax.get_as_vector(), 10 * tau / c.Kmax.norm(), 255, 0, 0);
        surface.draw_flow(c.Kmin.get_as_vector(), 10 * tau / c.Kmin.norm(), 0, 255, 0);
      }
    if(parent.draw_normals) /* affiche les normales sur les faces */
      surface.draw_faces_normals(10);
  }
}

class SurfaceManager {
  ArrayList<SurfaceInfo> surface = new ArrayList<SurfaceInfo>();
  SurfaceInfo current;
  CurvatureEngine curvature;
  boolean draw_normals = false;
  boolean do_color = true;
  boolean draw_vector_field = true;
  boolean draw_principals_flow = true;
  AutoOffsetProxyGui gui = new AutoOffsetProxyGui(new PVector(25, 50));
  boolean hide_gui = false;
  int deform_count = 0;
  
  SelectorGuiComponent curvature_select = new SelectorGuiComponent("Curvature", 12.5);
  SelectorGuiComponent file_select = new SelectorGuiComponent("File", 12.5);
  
  SurfaceManager(SurfaceData data, CurvatureEngine curvature) {
    /* Pour être plus efficace sur la récupération des voisins et faces nous utiliserons des listes d'adjacence */
    surface.add(new SurfaceInfo(new AdjacencyListSurface(data)));
    current = surface.get(0);
    this.curvature = curvature;
    
    build_gui();
    prepare_flow();
  }
  
  private void build_gui() {
    gui.addComponent(new BarGuiComponent("Tau", log(tau) / log(2), -5, 10, 12.5, 150, new Function<Float, Float>(){
      @Override
      public Float apply(Float val) {
        return pow(2, val);
      }
    }));
    gui.addComponent(new CheckBoxGuiComponent("Animate", 12.5, do_animation));
    gui.addComponent(new CheckBoxGuiComponent("A posteriori renormalization", 12.5, do_auto_scale));
    gui.addComponent(new CheckBoxGuiComponent("Draw flow", 12.5, draw_vector_field));
    //gui.addComponent(new CheckBoxGuiComponent("Draw principals flow", 12.5, draw_principals_flow));
    gui.addComponent(new CheckBoxGuiComponent("Draw normals", 12.5, draw_normals));
    gui.addComponent(new CheckBoxGuiComponent("Draw error as colors", 12.5, do_color));
    gui.addComponent(new CheckBoxGuiComponent("Draw axis", 12.5, draw_axis));
    gui.addComponent(new CheckBoxGuiComponent("Ground truth from start file", 12.5, ground_truth_from_start_data));
    gui.addComponent(new CheckBoxGuiComponent("Minimize", 12.5, minimize));
    
    gui.addComponent(curvature_select);
    
    file_select.current_id = -1;
    gui.addComponent(file_select);
    
    gui.refresh();
  }
  
  void addCurvature(String name, CurvatureEngine curvature) {
    curvature_select.add(name, curvature);
    
    gui.refresh();
  }
  
  void addFile(String name) {
    file_select.add(name, name);
    
    gui.refresh();
  }
  
  void reload(SurfaceData data) {
    surface.clear();
    surface.add(new SurfaceInfo(new AdjacencyListSurface(data)));
    current = surface.get(0);
    deform_count = 0;
    
    prepare_flow();
    prepare_flow(); /* solve strange bug with FacesCurvature ??? */
  }
  
  void addGroundTruth(SurfaceData data) {
    current.addGroundTruth((data == null) ? null : new AdjacencyListSurface(data));
  }
  
  void deform() {
    current.apply_flow(this, curvature, tau);
  }
  
  void auto_scale() {
    current.auto_scale();
  }
  
  void center() {
    current.center();
  }
  
  void scale(float scalar) {
    current.scale(scalar);
  }
  
  void prepare_flow() {
      current.prepare_flow(curvature);
  }
  
  void getValueFromGui(String name) {
    switch(name) {
    case "Tau":
      tau = (Float)gui.get(name);
      break;
      
    case "A posteriori renormalization":
      do_auto_scale = (Boolean)gui.get(name);
      break;
      
    case "Animate":
      do_animation = (Boolean)gui.get(name);
      break;
      
    case "Draw flow":
      draw_vector_field = (Boolean)gui.get(name);
      break;
    
    case "Draw principals flow":
      draw_principals_flow = (Boolean)gui.get(name);
      break;
      
    case "Draw normals":
      draw_normals = (Boolean)gui.get(name);
      break;
      
    case "Draw error as colors":
      do_color = (Boolean)gui.get(name);
      break;
    
    case "Draw axis":
      draw_axis = (Boolean)gui.get(name);
      break;
  
    case "Minimize":
      minimize = (Boolean)gui.get(name);
      break;
    
    case "Ground truth from start file":
      ground_truth_from_start_data = (Boolean)gui.get(name);
      addGroundTruth(ground_truth_from_start_data ? current.surface.get_data().copy() : null);
      break;
    
    case "Curvature":
      curvature = (CurvatureEngine)gui.get(name);
      prepare_flow();
      break;
    
    case "File":
      file_to_load = (String)gui.get(name);
      file_select.current_id = -1;
      break;
    }
  }
  
  void getValuesFromGui() {
    for(String e : gui.values())
      getValueFromGui(e);
  }
  
  void mousePressed() {
    if(! hide_gui)
      gui.mousePressed();
    if(gui.isInGui())
      getValueFromGui(gui.current().getName());
  }
  
  void mouseReleased() {
    if(gui.isInGui())
      getValueFromGui(gui.current().getName());
    gui.mouseReleased();
  }
  
  void draw_infos() {
    fill(127);
    if(! hide_gui) {
      gui.draw();
      if(gui.isInGui())
        getValueFromGui(gui.current().getName());
      text("Number of deformations : " + deform_count, 25, 25);
    }
    
  }
  
  void draw() {
    current.draw(this);
  }
}
