/* ---------------------------------------------------------------------------- */
/* Test                                                                         */
/* ---------------------------------------------------------------------------- */

/* Ensemble de méthodes globales permettant de tester des fonctionnalités       */
/* précises attendues.                                                          */
/* Utile si les fonctionnalités sont difficiles à tester à l'oeil ou pour avoir */
/* une garantie qu'une fonctionnalité est correcte dans l'application.          */

import java.util.function.Consumer;

/* Classe d'un test unitaire. */
class Test <D> {
  String name; /* Nom du test pour l'output. */
  D data; /* données sur lesquelles effectuer le test. */
  Consumer<D> test; /* méthode à lancer sur le test pour le vérifier. */
  Test(String name, D data, Consumer<D> test) {
    this.name = name;
    this.data = data;
    this.test = test;
  }
  
  void run() {
    try {
      test.accept(data);
    } catch (Exception e) {
      println("TEST (" + name + ") : FAILURE");
      throw e;
    }
    println("TEST (" + name + ") : OK");
  }
}

/* Ensemble de tests prédéfinis. */
/* Ici nous souhaitions vérifier la validité de la méthode OrientedNeighbors des surfaces proposées. */
/* Si le test passe, alors la liste des voisins est ordonnée comme attendu. */
/* Le second test est de vérifier dans le cas de AdjacencyListSurface que toutes les faces suivent la même orientation (surface orientée) */
void test_global(SurfaceData data) {
  ArrayList<Test<?>> tests = new ArrayList<Test<?>>(); /* ensemble des tests à réaliser */
  
  Consumer<Surface> test_oriented = new Consumer<Surface>() { /* méthode de vérification du test : appel à la méthode de test test_OrientedNeighbors */
    public void accept(Surface d) { /* Note : this java is too old to do an one-liner with a lambda ... */
      test_OrientedNeighbors(d);
    }
  };
  
  Consumer<AdjacencyListSurface> test_surface_oriented = new Consumer<AdjacencyListSurface>() { /* méthode de vérification du test : appel à la méthode de test test_FaceOrientation */
    public void accept(AdjacencyListSurface d) {
      test_OrientedNeighbors(d);
    }
  };
  
  /* ajout des tests sur les différentes types de surfaces connus. */
  tests.add(new Test("ClassicSurface neighbors ordered", new ClassicSurface(data), test_oriented));
  tests.add(new Test("AdjacencyListSurface neighbors ordered", new AdjacencyListSurface(data), test_oriented));
  tests.add(new Test("AdjacencyListSurface surface orientation", new AdjacencyListSurface(data), test_surface_oriented));
  
  for(Test<?> test : tests) /* évaluation de chaque test. */
    test.run();
}

/* En cas d'échec de la vérification lance une IllegalStateException, si c'est le cas vérifier le code de la méthode orientedNeighbors() dans la Surface. */
void test_OrientedNeighbors(Surface s) {
  SurfaceData data = s.get_data();
  ArrayList<IntList> verif = new ArrayList<IntList>();
  for(int i = 0; i < data.vertices.size(); ++i)
    verif.add(s.orientedNeighbors(i));
  for(IntList face : data.faces) { /* on vérifie pour toute face */
    for(int i = 0; i < face.size(); ++i) {
      IntList v = verif.get(face.get(i)); /* pour chaque sommet de cette face */
      for(int k = 0; k < v.size(); ++k) {
        if(v.get(k) != face.get((i + 1) % face.size())) /* on veut que pour une arête de la forme u-p-v, u et v soient suivants l'un de l'autre dans la liste des voisins de p */
          continue;
        int check = face.get((i - 1 + face.size()) % face.size());
        if(v.get((k + 1) % v.size()) != check && v.get((k - 1 + v.size()) % v.size()) != check) {
          throw new IllegalStateException("Invalid property check for the vertex " + face.get(i) + ", for u = " + v.get(k) + ", expected to find " + check + " in {" + v.get((k + 1) % v.size()) + ", " + v.get((k - 1 + v.size()) % v.size()) + "}");
        }
      }
    }
  }
}

void test_FaceOrientation(AdjacencyListSurface s) {
  SurfaceData data = s.get_data();
  for(int v_id = 0; v_id < data.vertices.size(); ++v_id) {
    edge_loop : for(Edge e : s.adjacency.get(v_id)) {
      IntList f1 = data.faces.get(e.faces.get(0));
      IntList f2 = data.faces.get(e.faces.get(1));
      for(int i1 = 0; i1 < f1.size(); ++i1) {
        for(int i2 = 0; i2 < f2.size(); ++i2) {
          if(f1.get(i1) != f2.get(i2))
            continue;
          if(f1.get((i1 + 1) % f1.size()) == f2.get((i2 + 1) % f2.size()))
            throw new IllegalStateException("Two adjacent faces have different orientation : " + e.faces.get(0) + " and " + e.faces.get(1) + " have both the edge (" + f1.get(i1) + ", " + f1.get((i1 + 1) % f1.size()) + ")");
          else
            continue edge_loop;
        }
      }
    }
  }
}
