
/*********************************************************************************************************
 Projet de géométrie différentielle discrète : flot par les faces en 3D
 
 Année 2018 - 2019
 
 TRANCHO Kevin      (M2 Sciences de l'image)
 VEILLERETTE Victor (M2 Sciences de l'image)
 
*********************************************************************************************************/

/*********************************************************************************************************
 À propos du travail proposé :
 
 Ce projet est réalisé dans le cadre du cours de géométrie différentielle discrète dispensé par 
 Pascal ROMON et Laurent HAUSWIRTH.
 
 L'objectif de ce projet est de proposer une visualisation de différentes applications de flots sur une
 surface en 3 dimensions depuis des flots de coubures donnés dans le cas discret.
 
 Pour ceci nous proposons de manipuler les surfaces depuis les données qui la caractérise : ses sommêts
 et ses faces depuis la classe 'SurfaceData' (Ce type de classe peut être obtenu pour le moment depuis
 le chargement d'un fichier au format Stanford (.ply)).
 Pour manipuler ces données nous proposons une surcouche par l'interface 'Surface' (une interface est un
 liste de méthodes attendues pour une classe dans le but de la manipuler de manière générique sur un
 comportement souhaité). L'utilisation de l'interface 'Surface' fait sens ici pour proposer deux manière
 de gérer les adjacences entre les sommêts par les classes suivantes :
  - Classicurface : gestion du graphe d'arête et des faces à l'aide d'une matrice d'adjacence (arêtes)
 et d'une matrice d'incidence (faces)
  - AdjacencyListSurface : gestion de la surface des listes d'adjacence (liste des voisins pour chaque
 sommêt donné). Pour cette 'Surface' nous proposons aussi de garder en mémoire les faces sur lesquelles
 se trouvent chaque arête pour remplacer la matrice d'incidence.
 Une classe AbstractSurface est proposée pour factoriser les comportements communs à ces deux structures
 de données : principalement les manipulations en interne des données 'SurfaceData', applications de flots,
 affichages de la surface et d'autres opérations qu'il n'est pas nécessaire de différentier.
 
 À noter que dans cette application nous avons choisi d'utiliser les listes d'incidence pour de meilleure
 performances sur les opérations de récupération de voisins et faces pour les différentes courbures.
 (La matrice d'adjacence permet de vérifier une arête en temps constant mais est linéaire en le nombre
 total de voisins pour récupérer la liste des voisins ce qui nous intéresse pour récupérer 'star(p)' dans
 des calculs de courbure moyenne. Alors que les listes d'adjacence sont linéaires en le nombre de voisins
 d'un sommêt ce qui nous semble plus intéressant pour les performances de nos calculs de courbures).
 
 Pour le calcul des flots nous proposons l'interface fonctionnelle 'CurvatureEngine' (ceci signifie que
 c'est une interface caractérisée par une seule méthode). Ceci laisse la possibilité d'être plus générique
 et proposer plusieurs courbures sous cette interface.
 Notons que nous attendons d'une courbure qu'elle prenne en entrée une surface et fabrique le champs de
 vecteurs (flot) à appliquer aux sommêts de cette surface pour la déformer par la courbure donnée.
 (Pour tout sommêt p de la surface et H(p) le vecteur issu du champs de vecteur correspondant à la
 déformation à appliquer en p, on renormalise H(p) par un paramètre de temps tau global pour toute la surface
 et on applique la déformation suivante : p' = p + tau * H(p) où p' est le sommet sur la nouvelle surface).
 À noter que le champs de vecteur est renormalisé par sa taille dans l'espace R^{3n} pour que la
 déformation des différentes courbures soit équivalente par rapport au tau choisi (règle des problèmes si
 les champs de vecteurs de deux coubures différentes ont des normes très éloignées l'un de l'autre).
 Nous proposons aussi de définir le flot par les arêtes depuis la classe abstraite 'EdgeCurvatureEngine'
 depuis laquelle hériter pour laquelle il suffit de redéfinir la méthode 'build_edge_vectors'.
 
 Les flots de courbures implémentés sont les suivantes:
  - Courbure harmonique (dans HarmonicCurvature.pde)
  - Courbure moyenne par les sommêts (dans MeanCurvature.pde)
  - Courbure moyenne par les arêtes (dans MeanCurvature.pde)
  - Courbure moyenne anisotropique par les arêtes (dans MeanCurvature.pde)
  - Courbure moyenne par les faces (dans MeanCurvature.pde)
  - Courbure normale par le gradient partiel sur le volume (GradientCurvaturedans CurvatureEngine.pde)
  - Courbures normale par les courbures principales (dans PrincipalsCurvatures.pde)
  - Courbures géodésique par les courbures principales (dans PrincipalsCurvatures.pde)
  
 Pour visualiser les concepts décrits précédemment, nous proposons une interface graphique pour gérer la
 surface étudiée ainsi que les différentes courbures dont le fonctionnement est plus détaillé dans le
 mannuel d'utilisation.
 Notons qu'il peut être intéressant de l'utiliser par exemple pour :
  - prévisuabliser les champs de vecteurs et comparer les courbures sur une même surface
  - visualiser les zones où les déformations sont les plus importantes (à l'aide des couleurs sur la surface)
  - Comparer à l'aide de la fonction ground truth la déformation depuis la surface d'origine
  - Regarder la réduction de bruit et lissage de la surface (par exemple à l'aide de bunny_[...]_noise.ply)
  - Alterner différentes courbures et observer les résultats
 
*********************************************************************************************************/

/*********************************************************************************************************
 Mannuel d'utilisation :
 
 Fonctionne sous Processing 3.4 ou supérieur.
 
 Au lancement de l'application, une surface est chargée depuis un fichier au format Stanford (.ply).
 Le fichier ouvert par défaut peut être changé en modifiant la valeur de 'file_to_load'.
 Le choix du fichier peut être changé pendant l'execution depuis l'interface utilisateur (rubrique File).
 Si un nouveau fichier doit être ajouté, il est possible d'indiquer son chemin dans le dossier 'data' en
 ajoutant la ligne "manager.addFile("mon_fichier.ply");" à la fin de la fonction 'setup'.
 Il est indiqué à l'écran le nombre de déformations, cette indication est le nombre de fois où le flot a
 été appliqué à la surface.
 
 Au cours de l'application, la surface sera déformée par un calcul de flot obtenu par la méthode 'build'
 décrite par 'curvature'. (Il est possible de changer la courbure par défaut)
 De même il est possible de changer la déformation à appliquer pendant l'execution en choissant une autre
 courbure depuis l'interface utilisateur (rubrique curvature).
 Pour ajouter une courbure à l'interface, il suffit d'ajouter à la fin de la fonction 'setup' la ligne:
 manager.addCurvature("Nom de ma courbure", new MaCourbure());
 À noter que 'MaCourbure' doit implémenter l'interface CurvatureEngine (respect du contrat donné par cette
 interface en implémentant les méthodes demandées : ici la méthode 'build')
 
 Le déplacement de la surface étudiée se fait à l'aide de la souris:
 
  - maintenir le clic pour déplacer la caméra autours de la surface
  - molette vers l'avant pour se rapprocher de la surface
  - molette vers l'arrière pour s'éloigner de la surface
 
 Une interface utilisateur (Gui) est proposée : ensemble de boutons cliquables à la souris sur l'écran.
 
  - tau : paramètre de temps utilisé pour la déformation : poids du champs du vecteurs pour déformer la 
 surface.
  - animation : permet de lancer ou arêter l'animation de déformation de la surface (Il est possible de
 laisser tourner l'animation et changer les paramètres en cours d'affichage).
  - renormalisation à posteriori : active ou non la renormalisation de la surface après application du flot
 (ici ce n'est pas une renormalisation sur l'aire ou le volume de la surface mais sur la taille du plus
 grand vecteur pour garder la surface affichée dans une taille équivalente à la précédente sur l'écran).
  - affichage du flot : affiche ou non le champs de vecteur définissant le flot (à noter qu'il est possible
 de changer la courbure ce qui changera la prévisualisation du flot).
  - affichage des normales : affiche ou non les normales sur les faces de la surface.
  - affichage des erreurs en tant que couleur : affiche sous forme de couleurs la distance soit à un ground
 truth, soit à la surface précédente. Le rouge correspond à la plus grande erreur (plus grande distance à
 la surface précédente/ground truth) et le vert à la plus petite erreur / distance.
  - affichage des axes : affiche ou non les axes (x, y, z) à l'écran.
  - Utiliser le fichier d'origine comme ground truth : laisse la possibilité de garder la surface chargée
 initialement comme surface de référence pour comparer la distance à celle-ci. Si cette option est inactive
 c'est la distance à la surface calculée au temps précédent qui sera affichée sous forme de couleur.
  - courbure : permet de sélectionner une courbure depuis les courbures pré-ajoutées dans 'setup'.
  - fichier : permet de charger ou recharger un fichier de surface depuis la liste de fichier pré-ajoutés.
 
 Il est aussi possible d'utiliser quelques raccourcis clavier bien qu'il soit préférable d'utiliser
 l'interface proposée. Les raccourcis intéressants sont par exemple:
  - 'h' hide : cache l'interface utilisateur pour observer seulement la surface.
  - 'a' animate : lance ou arête l'animation (utile principalement si l'interface est cachée)
 
 Il est aussi possible d'utiliser une surface de référence si elle est connue (par exemple si on bruité ou
 modifié une surface connue à l'avance):
 Pour ceci il est possible de regarder la fonctionnalité 'Ground Truth connu' dans 'setup'.
 
 La 'courbure par les arêtes avec la méthode de Hildebrandt et Polthier' est aussi implémentée, mais elle
 n'est pas ajoutée à l'interface. Il est possible de jouer sur ses paramètres et la tester en regardant
 la 'Fonctionnalité courbure par les arêtes avec la méthode de Hildebrandt et Polthier' dans 'setup'.
 
*********************************************************************************************************/

/*********************************************************************************************************
 À propos de l'architecture / modules du projet :
 
 ~ Surfaces ~
 
  - Surface : interface des comportements attendus pour être manipulé comme une surface
  - SurfaceData : données de la surface : ensemble des sommêts et des faces
  - AbstractSurface : méthodes de manipulation des surfaces (affichage, flots, gestion des données)
  - ClassicSurface : implémentation du voisinage par matrice d'adjacence et matrice d'incidence
  - AdjacencyListSurface : implémentation du voisinage par listes d'adjacence
 
 ~ Courbures ~ 
 
  - CurvatureEngine : interface des comportements attendus pour être utilisé comme une recette de fabrication
  du flot
    - EdgeCurvatureEngine : classe abstraite pour définir un flot par les arêtes
  - HarmonicCurvature : courbure harmonique sur les surfaces
  - VerticesMeanCurvature : courbure moyenne sur les surfaces
  - EdgesMeanCurvature : courbure moyenne sur les surfaces par les arêtes (classique et anisotropique)
  - FacesMeanCurvature : courbure moyenne sur les surfaces par les faces
  - PrincipalsCurvatures : courbures principales
 
 ~ Interface utilisateur ~
 
  - SurfaceManager : Gestionnaire des surfaces et des courbures (interface utilisateur)
  - Gui : modèle de construction des interfaces graphique utilisateur
  - GuiComponent : interface d'un composant et composants pour le Gui
  - Camera : camera TrackBall
 
 ~ Outils ~
 
  - VectorField : interface définissant des opérations à respecter pour être un objet mathématiques plus générique
  - PVectorList : liste de PVector avec quelques fonctionnalités intéressantes pour être manipulées comme un
  vecteur dans R^{3n} (est un VectorField)
  - PMatrix : matrices 3x3 pour les calculs de valeurs et vecteurs propres
  - Tools : quelques outils comme l'algorithme de Gram-Schmidt
  - Tests : quelques tests unitaires pour vérifier le bon fonctionnement de certaines fonctionnalités
 
*********************************************************************************************************/

/*********************************************************************************************************
 Bugs identifiés à corriger, difficultés rencontrées toujours d'actualité :
 
 ### BUG : lors de changements de flots ou par défaut, le flot par les faces peut dérailler:
 Deux observations distinctes : dans certains cas semble OK sur les formes régulières mais des fois semble
 presque aléatoire. Cliquer sur un autre flot puis revenir règle des fois le problème.
 Semble venir du flot par les faces. (peut demander plusieurs chargements : écrasement puis flot ?)
 
*********************************************************************************************************/

/*********************************************************************************************************
 Notes des tâches pour les développeurs :
 
 /!\ Tâches prioritaires /!\ :
  
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
 ~~~ Tâches à faire ~~~ :
 
 ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
 ~~~ Améliorations à faire / possibles ~~~ :
 
 -Meilleure gestion de la caméra (la trackball caméra ne permet pas une rotation agréable de l'objet)
 Solution possible : utiliser des quaternions.
 -Gestion des contraintes (en accord avec l'architecture logicielle existante)
 -Comparaison de courbures -> plusieurs courbures simultanées ? Outils de comparaison ?
 -Courbure géodésique ?
 -Hildebrandt et Polthier : ajouter flot de courbure avec paramétrisation dans le Gui ?
 
*********************************************************************************************************/

String file_to_load = "icosphere_4_subdivs.ply"; /* Premier fichier ouvert lors du lancement de l'application */
CurvatureEngine curvature = new HarmonicCurvature(true); /* Première courbure sélectionnée, il est possible d'en changer à l'aide de l'interface utilisateur */

SurfaceManager manager;
Camera camera;

/* Paramètres par défaut au lancement de l'application : */

float tau = 25; /* paramètre de temps pour l'application de la déformation par la courbure */
boolean do_tests = true; /* faire les tests de vérification au lancement de l'application */
boolean do_auto_scale = true; /* paramètre par défaut de la renormalisation a posteriori */
boolean do_animation = false; /* l'animation ne se lance pas au lancement de l'application */
boolean draw_axis = false; /* par défaut les axes ne sont pas affichés */
boolean ground_truth_from_start_data = false; /* désactive la comparaison frame par frame pour une comparaison au fichier d'origine */
boolean minimize = true; /* réduit la surface par le flot */

float auto_scale_size; /* taille de la renormalisation à posteriori, sera par défaut donnée dans setup */

boolean other_style = false; /* change le design (les couleurs) de l'application */
boolean lines_on_surface = false; /* affiche ou non les arêtes sous forme de lignes sur le mesh */
boolean grid = false; /* ne dessine pas les faces */

void setup() {
  size(1600, 900, P3D);
  frameRate(60);
  
  camera = new Camera(width / 2);
  
  auto_scale_size = width / 4;
  
  if(do_tests) /* tests unitaires de quelques fonctionnalités de l'application */
    test_global(loadPLY("bunny.ply"));
    
  SurfaceData original_data = loadPLY(file_to_load);
  file_to_load = null;
  
  manager = new SurfaceManager(original_data, curvature);
  
  /* *** Fonctionnalité Ground Truth connu : *** */
  
  /* Il est possible de comparer le résultat de la courbure à une surface de référence */
  /* Mais cette fonctionnalité n'est pas encore ajoutée à l'interface utilisateur */
  /* Il est possible de l'activer en décommentant les lignes suivantes : */
  //SurfaceData ground_truth = loadPLY("bunny.ply");
  //manager.addGroundTruth(ground_truth);
  
  manager.auto_scale();
  manager.center();
  
  /* Liste des fichiers proposés pour tester l'application : */
  
  manager.addFile("bunny.ply");
  //manager.addFile("bunny_weak_noise.ply");
  //manager.addFile("bunny_noise.ply");
  //manager.addFile("bunny_strong_noise.ply");
  manager.addFile("bunny_excessive_noise.ply");
  manager.addFile("bunny_test_excessive_noise.ply");
  manager.addFile("mug.ply");
  manager.addFile("pyramid.ply");
  manager.addFile("cube.ply");
  //manager.addFile("cube_triangulated_subdiv.ply");
  manager.addFile("biased_geodesic.ply");
  manager.addFile("geodesic.ply");
  manager.addFile("dodecahedron.ply");
  manager.addFile("torus.ply");
  manager.addFile("double_torus.ply");
  manager.addFile("multi_torus.ply");
  manager.addFile("icosphere_4_subdivs.ply");
  manager.addFile("icosphere_singularities.ply");
  manager.addFile("sphere10-20.ply");
  //manager.addFile("bunny_subdiv1_noisy.ply");
  
  /* Liste des courbures proposées pour tester l'application */
  
  manager.addCurvature("Harmonic curvature", new HarmonicCurvature());
  manager.addCurvature("Harmonic curvature (probabilistic)", new HarmonicCurvature(true));
  manager.addCurvature("Face curvature", new FacesCurvature());
  manager.addCurvature("Mean curvature by the vertices", new VerticesMeanCurvature());
  manager.addCurvature("Mean curvature by the vertices (probabilistic)", new VerticesMeanCurvature(true));
  manager.addCurvature("Mean curvature by the edges", new EdgesMeanCurvature());
  manager.addCurvature("Mean curvature by the faces", new FacesMeanCurvature());
  manager.addCurvature("Partial gradient normal curvature", new GradientCurvature());
  manager.addCurvature("Normal curvature", new PrincipalNormalCurvature());
  manager.addCurvature("Gaussian curvature", new GaussianCurvature());
  
  /* *** Fonctionnalité courbure par les arêtes avec la méthode de Hildebrandt et Polthier (lambda, r) : *** */
  
  /* La courbure suivante n'est pas ajoutée à l'interface mais il est possible de la tester */
  /* et jouer sur les paramètres en décommentant la ligne suivante : */
  //manager.addCurvature("HildebrandtPolthierEdgesMeanCurvature", HildebrandtPolthierEdgesMeanCurvature(1, 1)); 
  
}

void draw() {
  try {
    
  if(file_to_load != null) { /* recharge un fichier si l'utilisateur en choisi un dans la liste des fichiers proposés */
    manager.reload(loadPLY(file_to_load));
    
    /* Il est possible de prendre comme référence le fichier d'origine et comparer */
    /* la distance des déformations successives à la surface d'origine. */
    if(ground_truth_from_start_data)
      manager.addGroundTruth(loadPLY(file_to_load));
    
    manager.auto_scale();
    manager.center();
    //manager.prepare_flow();
    file_to_load = null;
  }
  if(other_style)
    background(255);
  else
    background(0x19, 0x20, 0x2E);

  manager.draw_infos(); /* Affichage d'informations en 2D et de l'interface utilisateur */
  
  pushMatrix();
  
  ambientLight(63, 63, 63, width, -width, width);
  pointLight(255, 255, 255, width, -width, width);
  
  camera.look(); /* gestion des mouvements de la caméra */
  
  // axes
  if(draw_axis) {
    line(0, 0, 0, width/4, 0, 0);
    line(0, 0, 0, 0, width/4, 0);
    line(0, 0, 0, 0, 0, width/4);
  }
  if(do_animation)
    manager.deform(); /* Déformation de la surface depuis la courbure choisie */
  if(do_auto_scale) {
    manager.auto_scale(); /* renormalisation à posteriori */
    manager.center();
  }
  manager.draw();
  
  popMatrix();
  
  } catch (Exception e) {
    e.printStackTrace();
    exit();
  }
}

void mouseWheel(MouseEvent event) {  // for zooming in and out
  float e = event.getCount();
  if(e != 0)
    camera.zoom(e < 0 ? 0.8 : 1.25);
}

void mousePressed() {
  manager.mousePressed();
  if(! manager.gui.isInGui())
    camera.active();
}

void mouseReleased() {
  if(! manager.gui.isInGui())
    camera.unactive();
  manager.mouseReleased();
}

void keyReleased() {
  switch(key) {
    case 'a' : do_animation = ! do_animation; break; /* run or stop the animation */
    case 'h' : manager.hide_gui = ! manager.hide_gui; break; /* hide the gui or show it */
    case 'z' : grid = ! grid; break;
  }
}
